/********************************************************************************************
 *                                                                                          *
 * Copyright (C) 2017 Armin Felder, Dennis Beier                                            *
 * This file is part of RocketChatMobileEngine <https://git.fairkom.net/chat/fairchat>.     *
 *                                                                                          *
 * RocketChatMobileEngine is free software: you can redistribute it and/or modify           *
 * it under the terms of the GNU General Public License as published by                     *
 * the Free Software Foundation, either version 3 of the License, or                        *
 * (at your option) any later version.                                                      *
 *                                                                                          *
 * RocketChatMobileEngine is distributed in the hope that it will be useful,                *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of                           *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                            *
 * GNU General Public License for more details.                                             *
 *                                                                                          *
 * You should have received a copy of the GNU General Public License                        *
 * along with RocketChatMobileEngine. If not, see <http://www.gnu.org/licenses/>.           *
 *                                                                                          *
 ********************************************************************************************/



#include "fileservice.h"

FileService::FileService( RocketChatServerData *server )
{
    if ( server != nullptr ) {
        mStorage = PersistanceLayer::instance();
        mTempPath = QStandardPaths::writableLocation( QStandardPaths::TempLocation );
        mDocumentPath = QStandardPaths::writableLocation( QStandardPaths::DocumentsLocation );

        if ( !mDocumentPath.size() ) {
            mDocumentPath = mTempPath;
        }

        mImagepath = QStandardPaths::writableLocation( QStandardPaths::PicturesLocation );

        if ( !mImagepath.size() ) {
            mImagepath = mDocumentPath;
        }

        mVideoPath = QStandardPaths::writableLocation( QStandardPaths::MoviesLocation );

        if ( !mVideoPath.size() ) {
            mVideoPath = mDocumentPath;
        }

        mMusicPath = QStandardPaths::writableLocation( QStandardPaths::MusicLocation );

        if ( !mMusicPath.size() ) {
            mMusicPath = mDocumentPath;
        }

        mServer = server;
    }
}


/**
 * This method tries to find a file and decides by itself to load it from disk or server
 *
 * WARNING: If the file is already in the download process,
 *
 * @brief FileService::getFileRessource
 * @param request
 * @param pThen function that is execeuted when thre is some result
 * @return bool true if the file is already downloaded, then is not executed in this case
 */
bool FileService::getFileRessource( QSharedPointer< FileRequest > pRequest )
{

    if ( !pRequest.isNull() && pRequest->url.size() > 0 ) {
        if ( mCurrentDownloads.contains( pRequest->url ) ) {
            return true;
        }  else {
            auto tempFile = getFileFromCache( pRequest );

            if ( !tempFile.isNull() ) {
                pRequest->then( tempFile, pRequest->showInline );
            } else {
                getFileFromServer( pRequest );
            }
        }
    }

    return false;
}

QSharedPointer<TempFile> FileService::getFileFromCache( QSharedPointer< FileRequest > pRequest )
{
    if ( !pRequest.isNull() ) {
        QString baseUrl = mServer->getRestApi()->getBaseUrl();
        QString fullUrl = baseUrl + pRequest->url ;
        qDebug() << "file from cache requested" << fullUrl;
        QString storageEntry = mStorage->getFileCacheEntry( fullUrl );
        qDebug() << "cachehit:" << fullUrl << ":" << storageEntry;
        qDebug() << "storageEntry len " << storageEntry.size();

        if ( !storageEntry.size() ) {
            return QSharedPointer<TempFile> ( nullptr );
        } else {
            QFile cacheHit( storageEntry );

            if ( cacheHit.exists() ) {
                return QSharedPointer<TempFile> ( new TempFile( storageEntry, pRequest->url ) );
            } else {
                mStorage->removeFileCacheEntry( pRequest->url, storageEntry );
                return QSharedPointer<TempFile> ( nullptr );
            }
        }
    } else {
        return QSharedPointer<TempFile> ( nullptr );
    }
}

void FileService::getFileFromServer( QSharedPointer< FileRequest > pRequest )
{
    if ( Q_LIKELY( !pRequest.isNull() ) ) {
        QMutexLocker( &this->mDownloadMutex );
        QString baseUrl = mServer->getRestApi()->getBaseUrl();
        QString fullUrl = baseUrl + pRequest->url ;

        if ( !mCurrentDownloads.contains( fullUrl ) ) {
            RestRequestCallback success = [ = ]( QNetworkReply * pReply, QJsonObject pData, RestApi * api ) {
                Q_UNUSED( pData );
                Q_UNUSED( api );

                if ( pReply ) {
                    QByteArray content = pReply->readAll();
                    QMimeType mimetype = mMimeDb.mimeTypeForData( content );
                    QString location = determineLocation( pReply->url(), mimetype );
                    QUrl locationUrl( location );
                    QString dirPath = locationUrl.adjusted( QUrl::RemoveFilename ).toString();
                    QStringList parts = location.split( '/' );

                    QDir dir( dirPath );

                    if ( !dir.exists() ) {
                        QString tempPath = parts.first();

                        for ( int i = 1; i < parts.length() - 1; i++ ) {
                            tempPath.append( '/' + parts[i] );
                            QDir temp( tempPath );

                            if ( !temp.exists() ) {
                                temp.mkdir( tempPath );
                            }
                        }
                    }

                    if ( dir.exists() ) {

                        QFile file( location );
                        QString url = pReply->url().toString();

                        if ( file.exists() ) {
                            QString fileHash = getFileHash( &file );
                            QString dataHash = QCryptographicHash::hash( content, QCryptographicHash::Md5 );

                            if ( fileHash == dataHash ) {
                                QSharedPointer<TempFile> tempfile( new TempFile( location, url ) );
                                pRequest->then( tempfile, pRequest->showInline );

                                if ( location.size() ) {
                                    mStorage->addFileCacheEntry( url, location );
                                    mCurrentDownloads.remove( fullUrl );
                                }

                                return;
                            }
                        }

                        if ( file.open( QIODevice::ReadWrite ) ) {
                            int written = file.write( content );
                            qDebug() << "wrote byte to file" << written;
                            file.close();
                            QSharedPointer<TempFile> tempfile( new TempFile( location, url ) );
                            mStorage->addFileCacheEntry( url, location );
                            pRequest->then( tempfile, pRequest->showInline );

                        } else {
                            qWarning() << "Could not write file" + location;
                            qWarning() << file.errorString();
                        }

                        mCurrentDownloads.remove( fullUrl );
                    } else {
                        qWarning() << "folder does not exist: " << dirPath;
                    }
                }
            };
            RestRequestCallback error = [ = ]( QNetworkReply * reply, QJsonObject pData, RestApi * api ) {
                Q_UNUSED( reply );
                Q_UNUSED( pData );
                Q_UNUSED( api );
                qDebug() << "could not download file" << pRequest->url;
            };

            mCurrentDownloads.insert( fullUrl );
            RestApiRequest request = QSharedPointer<RestRequest>( new GetRequest( fullUrl, success, error, true, true ) );

            mServer->sendApiRequest( request );
        }
    }
}

QString FileService::getFileHash( QFile *pFile )
{
    QString hashString = "";

    if ( pFile ) {
        if ( pFile->open( QFile::ReadOnly ) ) {
            QCryptographicHash hash( QCryptographicHash::Md4 );

            if ( hash.addData( pFile ) ) {
                hashString =  hash.result();
            }

            pFile->close();
        }
    }

    return hashString;
}

QString FileService::determineLocation( QUrl pUrl, QMimeType pType )
{
    QString mimeTypeName = pType.name();
    QString urlString = pUrl.toString();
    bool image = mimeTypeName.contains( QStringLiteral( "image" ) );
    bool video = mimeTypeName.contains( QStringLiteral( "video" ) );
    bool audio = mimeTypeName.contains( QStringLiteral( "audio" ) );
    int hits = image + video + audio;
    QString urlPath = QStringLiteral( "/rocketChat" ) + pUrl.path();
    QString filePath;

    if ( hits > 1 ) {
        qDebug() << "More than one hit for one file" << urlString;
    }

    //Checks for wrongly named svg… please dont ask

    if ( image ) {
        filePath = mImagepath + urlPath;
    } else if ( video ) {
        filePath = mVideoPath + urlPath;
    } else if ( audio ) {
        filePath = mMusicPath + urlPath;
    } else {
        filePath = mTempPath + urlPath;
    }

    if ( pType.name().contains( "svg" ) ) {
        QStringList parts = filePath.split( "." );

        //change ending to correct ending if mimetype and ending collide
        if ( !parts.isEmpty() && parts.last() == "jpg" ) {
            QString origSuffix = parts.last();
            QString suffix = pType.preferredSuffix();
            qDebug() << "origSuffix " << origSuffix;
            int origSuffixLength = origSuffix.length() + 1;
            filePath.replace( filePath.length() - origSuffixLength, origSuffixLength, "." + suffix );
        }
    }

    return filePath;
}

FileRequest::FileRequest( QString url, QString type,  std::function<void ( QSharedPointer<TempFile>, bool )> then, bool showInline )
{
    this->type = type;
    this->url = url;
    this->then = then;
    this->showInline = showInline;
}
