/********************************************************************************************
 *                                                                                          *
 * Copyright (C) 2017 Armin Felder, Dennis Beier                                            *
 * This file is part of RocketChatMobileEngine <https://git.fairkom.net/chat/fairchat>.     *
 *                                                                                          *
 * RocketChatMobileEngine is free software: you can redistribute it and/or modify           *
 * it under the terms of the GNU General Public License as published by                     *
 * the Free Software Foundation, either version 3 of the License, or                        *
 * (at your option) any later version.                                                      *
 *                                                                                          *
 * RocketChatMobileEngine is distributed in the hope that it will be useful,                *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of                           *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                            *
 * GNU General Public License for more details.                                             *
 *                                                                                          *
 * You should have received a copy of the GNU General Public License                        *
 * along with RocketChatMobileEngine. If not, see <http://www.gnu.org/licenses/>.           *
 *                                                                                          *
 ********************************************************************************************/



#include "rocketchatchannelservice.h"



RocketChatChannelService::RocketChatChannelService( RocketChatServerData *pServer, MessageService *pMessageService )
{
    this->mServer = pServer;
    this->mMessageService = pMessageService;
    mStorage = PersistanceLayer::instance();
    qRegisterMetaType<channelVector>( "channelVector" );
    qRegisterMetaType<userVector>( "userVector" );

}

QSharedPointer<RocketChatChannel> RocketChatChannelService::createChannelObject( QString pRoomId, QString pName, QString pType )
{
    QSharedPointer<RocketChatChannel> ptr( new RocketChatChannel( mServer, mMessageService, pRoomId, pName, pType ) );

    if ( Q_LIKELY( mChannels ) ) {
        mChannels->add( ptr );
    } else {
        qCritical() << "Channelsrepo not available to ChannelService";
    }

    fillChannelWithMessages( ptr );

    return ptr;
}


QVector<QSharedPointer<RocketChatChannel> > RocketChatChannelService::processChannelData( QJsonArray pChannelArray, bool pJoined )
{
    QVector<QSharedPointer<RocketChatChannel> > vec;

    std::tuple<QString, QString> openChannelTupel =  mStorage->getCurrentChannel();
    QString openChannel = std::get<0>( openChannelTupel );
    mStorage->transaction();

    for ( const auto &currentChannel : pChannelArray ) {
        QJsonObject currentChannelObject = currentChannel.toObject();

        if ( Q_LIKELY( ( currentChannelObject.contains( "rid" ) &&
                         currentChannelObject.contains( "name" ) ) || ( currentChannelObject.contains( "_id" ) &&
                                 currentChannelObject.contains( "t" ) ) ) ) {

            QString id = currentChannelObject["rid"].toString();
            QString name = currentChannelObject["name"].toString();
            QString type = currentChannelObject["t"].toString();

            if ( id == "" ) {
                id = currentChannelObject["_id"].toString();
            }

            QSharedPointer<RocketChatChannel> channel = nullptr;

            if ( mChannels->contains( id ) ) {
                channel = mChannels->get( id );
            } else {
                channel = createChannelObject( id, name, type );
            }

            if ( !channel.isNull() ) {


                if ( !name.isEmpty() ) {
                    channel->setName( currentChannelObject["name"].toString() );
                }

                if ( currentChannelObject.contains( "ro" ) ) {
                    channel->setReadOnly( currentChannelObject["ro"].toBool() );
                }

                if ( currentChannelObject.contains( "archived" ) ) {
                    channel->setArchived( currentChannelObject["archived"].toBool() );
                }

                if ( currentChannelObject.contains( "blocker" ) ) {
                    channel->setBlocked( currentChannelObject["blocker"].toBool() );

                    if ( currentChannelObject["blocker"].toBool() ) {
                        mMessageService->addChannelToBlockList( channel->getRoomId() );
                    }
                }

                if ( currentChannelObject.contains( "muted" ) ) {
                    auto muted = currentChannelObject["muted"].toArray().toVariantList();
                    auto mutedUsers = QVariant( muted ).toStringList();
                    channel->setMuted( mutedUsers );

                    if ( currentChannelObject.contains( "username" ) && mutedUsers.contains( currentChannelObject["username"].toString() ) ) {
                        channel->setSelfMuted( true );
                    }
                }

                if ( id == openChannel ) {
                    channel->setUnreadMessages( 0 );
                } else if ( currentChannelObject.contains( "unread" ) ) {
                    channel->setUnreadMessages( currentChannelObject["unread"].toInt() );
                }

                channel->setJoined( pJoined );
                loadUsersOfChannel( channel );


                if ( channel->getArchived() ) {
                    subscribeChannel( channel );
                }


                vec.append( channel );
                persistChannel( channel );
            }

        }
    }

    mStorage->commit();


    return vec;
}


void RocketChatChannelService::loadJoinedChannelsFromServer( void )
{
    std::function<void ( QJsonObject, MeteorDDP * )> subscriptionSuccess = [ = ]( QJsonObject pResponse, MeteorDDP * pDdpApi ) {
        Q_UNUSED( pDdpApi );

        if ( Q_LIKELY( pResponse.contains( "result" ) ) ) {
            QJsonArray channels = pResponse["result"].toArray();
            auto channelsParsed = processChannelData( channels, true );
            emit channelsLoaded( channelsParsed, true );
        }
    };
    QSharedPointer<DDPRequest> subscriptions( new DDPMethodRequest( "subscriptions/get", {}, [ = ]( QJsonObject pResponse, MeteorDDP * pDdpApi ) {
        subscriptionSuccess( pResponse, pDdpApi );
        QSharedPointer<DDPRequest> rooms( new DDPMethodRequest( "rooms/get", {}, subscriptionSuccess ) );
        mServer->sendDdprequest( rooms );
    } ) );
    mServer->sendDdprequest( subscriptions );
}



void RocketChatChannelService::loadJoinedChannelsFromDb()
{
    QList<QVariantHash> roomsList = mStorage->getChannels();

    if ( Q_LIKELY( !roomsList.isEmpty() ) ) {
        QVector<QSharedPointer<RocketChatChannel>> channels;

        for ( const auto &roomHash : roomsList ) {
            QSharedPointer<RocketChatChannel> channelPointer =  createChannelObject( roomHash["id"].toString(), roomHash["name"].toString(), roomHash["type"].toString() );

            if ( !channelPointer.isNull() ) {
                channelPointer->setMuted( roomHash["list"].toStringList() );
                channelPointer->setArchived( roomHash["archived"].toBool() );
                channelPointer->setReadOnly( roomHash["readOnly"].toBool() );
                channelPointer->setJoined( roomHash["joined"].toBool() );

                if ( roomHash.contains( "blocked" ) ) {
                    channelPointer->setBlocked( roomHash["blocked"].toBool() );
                }

                channels.append( channelPointer );
            }
        }

        emit  channelsLoaded( channels, true );
    }
}

void RocketChatChannelService::persistChannel( QSharedPointer<RocketChatChannel> pChannel )
{
    QString id = pChannel->getRoomId();
    QString name  = pChannel->getName();
    QString type = pChannel->getType();
    QString muted = pChannel->getMuted().join( "," );
    bool joined = pChannel->getJoined();
    bool archived = pChannel->getArchived();
    bool readonly = pChannel->getReadOnly();
    bool blocked = pChannel->getBlocked();
    mStorage->addChannel( id, name, type, joined, readonly, muted, archived, blocked );
}
/**
 * Fills the channel with users, which are loaded from the server
 *
 * @brief RocketChatChannelService::loadUsersOfChannel
 * @param channel
 */
void RocketChatChannelService::loadUsersOfChannel( QSharedPointer<RocketChatChannel> pChannel )
{
    if ( !pChannel.isNull() ) {
        QString channelId = pChannel->getRoomId();
        QJsonArray params = {channelId, true};
        QSharedPointer<DDPMethodRequest> request( new DDPMethodRequest( "getUsersOfRoom", params ) );
        std::function<void ( QJsonObject, MeteorDDP * )> success = [ = ]( QJsonObject data, MeteorDDP * ddp ) {
            Q_UNUSED( ddp );

            if ( Q_LIKELY( data.contains( "result" ) ) ) {
                QJsonObject result = data["result"].toObject();

                if ( Q_LIKELY( result.contains( "records" ) && result["records"].isArray() ) ) {
                    QJsonArray records = result["records"].toArray();
                    QVector<QSharedPointer<RocketChatUser>> userList;

                    //crate new user objects and add them to the channel
                    for ( const auto &current : records ) {
                        if ( current.type() == QJsonValue::Object ) {
                            QJsonObject userObject = current.toObject();
                            QString name = userObject["username"].toString();
                            QSharedPointer<RocketChatUser> user( new RocketChatUser( name ) );
                            userList.append( user );
                        } else if ( current.type() == QJsonValue::String ) {
                            QString name = current.toString();
                            QSharedPointer<RocketChatUser> user( new RocketChatUser( name ) );
                            userList.append( user );
                        }
                    }

                    emit usersLoaded( channelId, userList );
                }
            }
        };
        request->setSuccess( success );
        mServer->sendDdprequest( request );
    }
}

void RocketChatChannelService::subscribeChannel( QSharedPointer<RocketChatChannel> pChannel )
{
    if ( !pChannel.isNull() ) {
        QString roomId = pChannel->getRoomId();
        QSharedPointer<RocketChatSubscribeChannelRequest> subscription( new RocketChatSubscribeChannelRequest( pChannel->getRoomId() ) );
        mServer->sendDdprequest( subscription );
    }
}



void RocketChatChannelService::fillChannelWithMessages( QSharedPointer<RocketChatChannel> pChannel )
{
    if ( !pChannel.isNull() ) {
        QString roomId = pChannel->getRoomId();
        auto messages = mStorage->getMessagesByRid( roomId );
        MessageList list;

        for ( const auto &currentMessage : messages ) {
            auto newMessage = mMessageService->parseMessage( currentMessage, false );

            if ( Q_LIKELY( !newMessage.isNull() ) ) {
                list.append( newMessage );
            }
        }

        pChannel->addMessages( list );
    }
}


RocketChatChannelService::~RocketChatChannelService()
{

}

MeteorDDP *RocketChatChannelService::getDdp() const
{
    return mDdp;
}

void RocketChatChannelService::setDdp( MeteorDDP *ddp )
{
    mDdp = ddp;
}

ChannelRepository *RocketChatChannelService::getChannels() const
{
    return mChannels;
}

void RocketChatChannelService::setChannels( ChannelRepository *pChannels )
{
    mChannels = pChannels;
}

void RocketChatChannelService::leaveChannel( const QString &pId )
{
    //TODO: handle response, as soon as RC is fixed to do so
    if ( mChannels->contains( pId ) ) {
        QSharedPointer<RocketChatLeaveRoomRequest> request( new RocketChatLeaveRoomRequest( pId ) );
        auto ddp = getDdp();
        ddp->sendRequest( request );
        deleteChannel( pId );
    }
}

void RocketChatChannelService::deleteChannel( const QString &pId )
{
    if ( mChannels->contains( pId ) ) {
        mChannels->get( pId )->setDeleted( true );
        mStorage->transaction();
        mStorage->deleteChannel( pId );
        mStorage->commit();
    }
}

void RocketChatChannelService::hideRoom( const QString &pId )
{
    QSharedPointer<RocketChatHideRoomRequest> request( new RocketChatHideRoomRequest( pId ) );
    mDdp->sendRequest( request );
    deleteChannel( pId );
}

void RocketChatChannelService::reportContent( const QString &pMessageId, const QString &pAuthorId )
{
    QJsonArray params = {QStringLiteral( "abuse" )};
    QSharedPointer<DDPMethodRequest> request( new DDPMethodRequest( QStringLiteral( "createDirectMessage" ), params ) );
    std::function<void ( QJsonObject, MeteorDDP * )> success = [ = ]( QJsonObject pResponse, MeteorDDP * pDdp ) {
        if ( Q_LIKELY( pResponse.contains( QStringLiteral( "result" ) ) ) ) {
            QJsonObject result = pResponse[QStringLiteral( "result" )].toObject();

            if ( Q_LIKELY( result.contains( QStringLiteral( "rid" ) ) ) ) {
                QString rid = result[QStringLiteral( "rid" )].toString();
                QSharedPointer<DDPSubscriptionRequest> subRoom( new DDPSubscriptionRequest( QStringLiteral( "room" ), {"d" + QStringLiteral( "abuse" )} ) );
                QSharedPointer<RocketChatSubscribeChannelRequest> subMessages( new RocketChatSubscribeChannelRequest( rid ) );
                QSharedPointer<RocketChatChannel> channel = createChannelObject( rid, QStringLiteral( "abuse" ), "d" );

                if ( !channel.isNull() ) {

                    emit channelsLoaded( {channel}, true );
                    pDdp->sendRequest( subRoom );
                    pDdp->sendRequest( subMessages );
                    emit directChannelReady( rid );
                    QString message = QStringLiteral( "Report message with id:" ) + pMessageId + QStringLiteral( " from user with id:" ) + pAuthorId;
                    QSharedPointer<RocketChatMessage> reportMessage( new RocketChatMessage( rid, message ) );
                    mMessageService->sendMessage( reportMessage );
                }
            }
        }

        pDdp->unsetResponseBinding( request->getFrame() );

    };
    request->setSuccess( success );
    mServer->sendDdprequest( request );
}

void RocketChatChannelService::openPrivateChannelWith( QString pUsername )
{
    QJsonArray params = {pUsername};
    QSharedPointer<DDPMethodRequest> request( new DDPMethodRequest( QStringLiteral( "createDirectMessage" ), params ) );
    std::function<void ( QJsonObject, MeteorDDP * )> success = [ = ]( QJsonObject pResponse, MeteorDDP * pDdp ) {
        if ( Q_LIKELY( pResponse.contains( QStringLiteral( "result" ) ) ) ) {
            QJsonObject result = pResponse[QStringLiteral( "result" )].toObject();

            if ( Q_LIKELY( result.contains( QStringLiteral( "rid" ) ) ) ) {
                QString rid = result[QStringLiteral( "rid" )].toString();
                QSharedPointer<DDPSubscriptionRequest> subRoom( new DDPSubscriptionRequest( QStringLiteral( "room" ), {"d" + pUsername} ) );
                QSharedPointer<RocketChatSubscribeChannelRequest> subMessages( new RocketChatSubscribeChannelRequest( rid ) );
                QSharedPointer<RocketChatChannel> channel = createChannelObject( rid, pUsername, "d" );

                if ( !channel.isNull() ) {
                    emit channelsLoaded( {channel}, true );
                    pDdp->sendRequest( subRoom );
                    pDdp->sendRequest( subMessages );
                    emit directChannelReady( rid );
                }
            }
        }

        pDdp->unsetResponseBinding( request->getFrame() );

    };
    request->setSuccess( success );
    mServer->sendDdprequest( request );
}
